#define MAIN
#include "MicroGlut.h"
#include "VectorUtils4.h"
#include "LoadTGA.h"
#include "GL_utilities.h"
#include "PingPong.h"
#include <cmath>
#include <iostream>
#include "fps.h"

const int screenWidth = 1024, screenHeight = 1024;

GLuint BG_texture;
GLuint cascade_shader = 0, screen_shader = 0;
float base_length = 1.0 / float(screenHeight);

unsigned int vertexArrayObjID;

PingPong pingpong;
Timer timer;

void init(void)
{
	dumpInfo();

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	printError("GL inits");

	cascade_shader = loadShaders("main.vert", "main.frag");
	screen_shader = loadShaders("screen.vert", "screen.frag");
	pingpong = {screenWidth, screenHeight};

	glUseProgram(cascade_shader);
	glUniform2f(glGetUniformLocation(cascade_shader, "resolution"), screenWidth, screenHeight);
	glUniform1i(glGetUniformLocation(cascade_shader, "BG_texture"), 3); // Texture unit 3

	LoadTGATextureSimple("textures/grass.tga", &BG_texture);
	glActiveTexture(GL_TEXTURE3);
	glBindTexture(GL_TEXTURE_2D, BG_texture);

	glGenVertexArrays(1, &vertexArrayObjID);
	glBindVertexArray(vertexArrayObjID);

	printError("init shader");

	int nr_layer = int(std::log(float(screenHeight)) / std::log(2.0)) + 1;

	for (int layer_index = nr_layer; layer_index > 0; layer_index--)
	{
		vec2 probe_size = vec2(screenHeight * pow(0.5, nr_layer - layer_index), screenWidth * pow(0.5, nr_layer - layer_index));
		std::cout << "layer index: " << layer_index << std::endl;
		std::cout << "probe size: " << probe_size.x << " " << probe_size.y << std::endl;
	    std::cout << "range: " << base_length * pow(4, layer_index - 1) << " " << base_length * pow(4.0, layer_index - 2) << std::endl ;
		std::cout << std::endl;
	}
}

void display(void)
{
	// Activate shader program
	glUseProgram(cascade_shader);

	float t = glutGet(GLUT_ELAPSED_TIME) / 1000.0;
	glUniform1f(glGetUniformLocation(cascade_shader, "t"), t);
	glUniform1f(glGetUniformLocation(cascade_shader, "base_length"), base_length);
	glUniform1i(glGetUniformLocation(cascade_shader, "texUnit"), 0);

	int nr_layer = int(std::log(float(screenHeight)) / std::log(2.0)) + 1;

	for (int layer_index = nr_layer; layer_index > 0; layer_index--)
	{
		// ray starts outside screenspace
		if (base_length * pow(4.0, layer_index - 2) > std::sqrt(2)) continue;

		vec2 probe_size = vec2(screenHeight * pow(0.5, nr_layer - layer_index), screenWidth * pow(0.5, nr_layer - layer_index));
		pingpong.swap();

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glUniform2f(glGetUniformLocation(cascade_shader, "probe_size"), probe_size.x, probe_size.y);
		glUniform1i(glGetUniformLocation(cascade_shader, "layer_index"), layer_index);
		glDrawArrays(GL_TRIANGLES, 0, 3);
	}

	glUseProgram(screen_shader);
	pingpong.draw_to_screen();

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glUniform1i(glGetUniformLocation(screen_shader, "texUnit"), 0);
	glUniform2f(glGetUniformLocation(screen_shader, "resolution"), screenWidth, screenHeight);
	glDrawArrays(GL_TRIANGLES, 0, 3);

	pingpong.clear_buffers();

	printError("display");
	glutSwapBuffers();
	timer.measure();
}

void mouse_func(int x, int y)
{
	glUseProgram(cascade_shader);
	glUniform2f(glGetUniformLocation(cascade_shader, "mouse_pos"), float(x) / float(screenWidth), 1.0 - float(y) / float(screenHeight));
}

void Key(unsigned char key,
		 __attribute__((unused)) int x,
		 __attribute__((unused)) int y)
{
	switch (key)
	{
	case '+':
		base_length *= 2.0;
		printf("base_length = %f\n", base_length);
		break;
	case '-':
		if (base_length > 0.0002)
			base_length *= 0.5;
		printf("base_length = %f\n", base_length);
		break;
	case 0x1b:
		exit(0);
	}
}

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);

	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
	glutInitWindowSize(screenWidth, screenHeight);

	glutInitContextVersion(3, 2);
	glutCreateWindow("Minimal Example");
	glutDisplayFunc(display);
	glutRepeatingTimer(0);

	glutKeyboardFunc(Key);
	glutPassiveMotionFunc(mouse_func);

	init();
	timer = Timer{50};
	glutMainLoop();
	exit(0);
}
