#version 430 core

#define PI 3.1415926538
	
uniform sampler2D texUnit;
uniform sampler2D BG_texture;
uniform float t;
uniform vec2 probe_size;
uniform float base_length;
uniform vec2 resolution;
uniform vec2 mouse_pos;
uniform int layer_index;

in vec2 texCoord;
out vec4 fragColor;

float eps = 0.001;
int max_iterations = 256;


struct Ray{

    vec2 pos;
    vec2 dir;
    float max_dis;
};

float sdBox( in vec2 p, in vec2 b )
{
    vec2 d = abs(p)-b;
    return length(max(d,0.0)) + min(max(d.x,d.y),0.0);
}

mat2 R(float t)
{
    return mat2(cos(t), -sin(t), sin(t), cos(t));        
}

float find_distance(vec2 pos){

    vec2 circle_pos = vec2(0.5, 0.5);
    vec2 circle2_pos = vec2(0.75, 0.75);// + vec2(0.2*sin(t), 0.2*cos(t));
    float circle_radius = 0.1;

    // float l1 = length(circle_pos - pos) - circle_radius;
    float l1 = sdBox((circle_pos - pos), vec2(0.2, 0.01));
    float l2 = length(circle2_pos - pos) - 0.2*circle_radius;
    float l3 = length(mouse_pos - pos) - 0.2*circle_radius;

    return min(l1, min(l2, l3));
}

vec4 get_color(vec2 pos){

    vec2 circle_pos = vec2(0.5, 0.5);// + vec2(0.2 * sin(t), 0);
    vec2 circle2_pos = vec2(0.75, 0.75);// + vec2(0.2*sin(t), 0.2*cos(t));
    float circle_radius = 0.1;

    float l1 = sdBox((circle_pos - pos), vec2(0.2, 0.01));
    // float l1 = length(circle_pos - pos) - circle_radius;
    float l2 = length(circle2_pos - pos) - 0.2*circle_radius;
    float l3 = length(mouse_pos - pos) - 0.2*circle_radius;

    if (l1 < eps){return vec4(0.0, 0.0, 0.0, 0.0);}
    if (l2 < eps){return 5*vec4(0.1, 0.1, 0.9, 0.0);}
    if (l3 < eps){return 5*vec4(0.8, 0.3, 0.1, 0.0);}
}

int is_out_of_bounds(vec2 t)
{
    return 1 - int(t.x < 0 || t.x >= 1 || t.y < 0 || t.y >= 1);        
}

vec4 ray_march(Ray ray){

    float t = 0.0;

    for(int i = 0; i < max_iterations; i++){
        float distance = 0.0;

        distance = find_distance(ray.pos);
        t += distance;

        if(distance < eps){
            return get_color(ray.pos);
        }
        if(t > ray.max_dis){
            return vec4(0.0, 0.0, 0.0, 1.0);
        }
        ray.pos += distance *ray.dir;
    }

    return vec4(0.0, 0.0, 0.0, 1.0);
}


vec4 get_far_color(vec2 pixel_coord, vec2 dir_coord, float dir_index, vec2 curr_probe_pos)
{    
    vec2 above_probe = probe_size * 2;
    vec2 probe_pos = floor(pixel_coord / above_probe) * above_probe;

    vec2 parent_probe_offset = vec2(
        curr_probe_pos.x > probe_pos.x ? -above_probe.x : above_probe.x,
        curr_probe_pos.y > probe_pos.y ? -above_probe.y : above_probe.y
    );

    vec2 base_probe_pos = probe_pos + dir_coord;

    vec2 pos1 = base_probe_pos / resolution;
    vec2 pos2 = (base_probe_pos + vec2(parent_probe_offset.x, 0)) / resolution;
    vec2 pos3 = (base_probe_pos + vec2(0, parent_probe_offset.y)) / resolution;
    vec2 pos4 = (base_probe_pos + parent_probe_offset) / resolution;

    float w1 = 0.35, w2 = 0.26, w3 = 0.13;
    
    vec4 color = vec4(0.0);
    color += w1 * is_out_of_bounds(pos1) * texture(texUnit, pos1);
    color += w2 * is_out_of_bounds(pos2) * texture(texUnit, pos2);
    color += w2 * is_out_of_bounds(pos3) * texture(texUnit, pos3);
    color += w3 * is_out_of_bounds(pos4) * texture(texUnit, pos4);

    return color;
}


void main()
{
    vec2 pixel_coord = texCoord * resolution;

    vec2 dir_coord = vec2(
        mod(int(pixel_coord.x), probe_size.x), 
        mod(int(pixel_coord.y), probe_size.y)
    );

    float dir_index = (dir_coord.x + dir_coord.y * probe_size.x)*4; 
    float dir_count = probe_size.x * probe_size.y*4;

    vec2 probe_pos = floor(pixel_coord / int(probe_size)) * probe_size + probe_size / 2;

    const float overlap_max = 1.2;
    const float overlap_min = 0.8;

    float max_t = overlap_max * base_length * pow(4, layer_index - 1);
    float min_t = overlap_min * layer_index > 1 ? base_length * pow(4.0, layer_index - 2) : 0.0;
    float max_dis = max_t - min_t;

    vec4 merged_color = vec4(0.0);
    bool hasMerged = false;

    for(int i=0; i < 4; i++){

        float new_dir_index = dir_index + i;
        
        vec2 dir_coord_offset = vec2(
            mod(new_dir_index, 2.0 * probe_size.y), 
            floor(new_dir_index / (2.0 * probe_size.x))
        );


        float angle = 2.0 * PI * ((new_dir_index + 0.5) / dir_count);
        vec2 dir = vec2(cos(angle), sin(angle));

        vec2 ray_start_pos = (probe_pos / resolution + dir * min_t);
        Ray r = Ray(ray_start_pos, dir, max_dis);

        vec4 ray_color = vec4(0.0);
        if(is_out_of_bounds(ray_start_pos) == 1){
            ray_color = ray_march(r);
        }

        vec4 far_color = vec4(0.0);
        if (ray_color.a == 1.0 && probe_size != resolution)
        {
            hasMerged = true;
            far_color = get_far_color(pixel_coord, dir_coord_offset, new_dir_index, probe_pos - probe_size / 2);  
        }
    
        merged_color += ray_color + far_color;  
    }

    merged_color = merged_color * 0.25;

    vec4 floor_color = mod(floor(10*texCoord.x) + floor(10*texCoord.y), 2) == 0 ? vec4(0.8) : vec4(0.2);
    // vec4 floor_color = texture(BG_texture, texCoord);
    floor_color = (hasMerged && layer_index == 1) ? floor_color : vec4(1);
    fragColor = floor_color * merged_color;
}
    
